typedef struct rc_entry_s rc_entry_t;

struct rc_entry_s {
	char key;
	char *prompt, *cmd;
	int ifkey, ifval;
	rc_entry_t *next;
};

rc_entry_t *rc = NULL;

#define KW_MAX 128
#define KW_MAXLEN 16
char keyword[KW_MAXLEN][KW_MAX];
char var[KW_MAXLEN];
int kwn = 0;

int kw_lookup(const char *str, int alloc)
{
	int n;

	for(n = 0; n < kwn; n++)
		if (strcmp(keyword[n], str) == 0)
			return n;

	if (!alloc || (kwn == KW_MAX) || (strlen(str) >= KW_MAXLEN)) return -1;

	n = kwn;
	strcpy(keyword[n], str);
	var[n] = -1;
	kwn++;
	return n;
}

void kw_set(char *cmd, int allow_get)
{
	char *end = strpbrk(cmd, " \t\r\n");
	int key, val;

	if (end == NULL) {
		if (!allow_get) {
			fd_printf(STO, "\r\n*** invalid set: missing value\r\n");
			return;
		}

		key = kw_lookup(cmd, 0);
		if (key < 0) {
			fd_printf(STO, "\r\n*** invalid key: %s\r\n", cmd);
			return;
		}
		fd_printf(STO, "\r\n*** %s=%s\r\n", keyword[key], var[key] < 0 ? "<unset>" : keyword[(int)var[key]]);
		return;
	}
	*end = '\0';
	end++;
	while(isspace(*end)) end++;
	key = kw_lookup(cmd, 0);
	if (key < 0) {
		fd_printf(STO, "\r\n*** invalid key: %s\r\n", cmd);
		return;
	}
	val = kw_lookup(end, 0);
	if (val < 0) {
		fd_printf(STO, "\r\n*** invalid value: %s\r\n", end);
		return;
	}
	var[key] = val;
}


#define RC_ERROR(msg) \
do { \
	fprintf(stderr, "rc syntax error in %s:%ld:\n  %s\n", fn, lineno, msg); \
	return; \
} while(0)

#define PREPEND_CURR() \
do { \
	curr->next = rc; \
	rc = curr; \
	curr = NULL; \
} while(0)


#define ALLOC_CURR(field, inval) \
do { \
	if ((curr != NULL) && (curr->field != inval)) PREPEND_CURR(); \
	if (curr == NULL) { \
		curr = calloc(sizeof(rc_entry_t), 1); \
		curr->ifkey = -1; \
	} \
} while(0)


static long rc_read_cmd(FILE *f, const char *stop, char *dst, long *lineno)
{
	long totlen = 0, len;
	char *end, *line, buf[4096];

	while((line = fgets(buf, sizeof(buf), f)) != 0) {
		(*lineno)++;
		end = strpbrk(line, "\r\n");
		if (end != NULL) {
			*end = '\0';
			if (strcmp(buf, stop) == 0)
				break;
			*end = '\n';
			len = end - buf + 1;
		}
		else
			len = strlen(line);
		totlen += len;
		if (dst != NULL) {
			memcpy(dst, buf, len);
			dst += len;
		}
	}

	if (dst != NULL)
		*dst = '\0';
	return totlen+1;
}

void rc_fparse(FILE *f, const char *fn)
{
	long lineno = 0;
	char *end, *arg, *line, buf[4096];
	rc_entry_t *curr = NULL;

	while((line = fgets(buf, sizeof(buf), f)) != 0) {
		lineno++;
		while(isspace(*line)) line++;
		if ((*line == '#') || (*line == '\0'))
			continue;
		end = strpbrk(line, "\r\n");
		if (end == NULL)
			RC_ERROR("Missing newline; line too long?");
		*end = '\0';
		arg = strpbrk(line, " \t");
		if (arg != NULL) {
			*arg = '\0';
			arg++;
		}
	
		if (strcmp(line, "key") == 0) {
			ALLOC_CURR(key, 0);
			if (*arg == '^') {
				arg++;
				if (isupper(*arg))
					*arg = *arg - 'A' + 1;
				else
					RC_ERROR("Invalid key; must be ^UPPERCASE");
			}
			curr->key = *arg;
		}
		else if (strcmp(line, "prompt") == 0) {
			int len = strlen(arg);
			ALLOC_CURR(prompt, 0);
			curr->prompt = malloc(len+1);
			memcpy(curr->prompt, arg, len+1);
		}
		else if (strcmp(line, "if") == 0) {
			ALLOC_CURR(ifkey, -1);
			end = strpbrk(arg, " \t\r\n");
			if (end == NULL)
				RC_ERROR("if requires two args");
			*end = '\0';
			end++;
			while(isspace(*end)) end++;
			curr->ifkey = kw_lookup(arg, 1);
			curr->ifval = kw_lookup(end, 1);
			if (curr->ifkey < 0)
				RC_ERROR("if: invalid var");
			if (curr->ifval < 0)
				RC_ERROR("if: invalid value");
		}
		else if (strcmp(line, "set") == 0) {
			kw_set(arg, 0);
		}
		else if (strcmp(line, "command") == 0) {
			long len = 0, pos = ftell(f);
			ALLOC_CURR(cmd, 0);
			len = rc_read_cmd(f, arg, NULL, &lineno);
			curr->cmd = malloc(len);
			fseek(f, pos, SEEK_SET);
			len = rc_read_cmd(f, arg, curr->cmd, &lineno);
		}
		else
			RC_ERROR("Unknown command");
	}

	if (curr != NULL)
		PREPEND_CURR();
}

void rc_parse(const char *fn)
{
	FILE *f = fopen(fn, "r");
	if (f == NULL)
		return;
	rc_fparse(f, fn);
	fclose(f);
}
